/*
 * This source file defines code for a program that solves the second XSEDE13 
 * Weekly Challenge problem.  A full description of the problem is available 
 * here: http://hpcuniversity.org/students/weeklyChallenge/56/ 
 *
 * The input to this program is a list of "test patterns."  A test pattern 
 * consists of a string of letters followed by a number.  The number indicates 
 * how many of the letters are "correct and in the correct position", i.e. how 
 * many letters in a solution exist in the same positions that they do in the 
 * test pattern.  Or, if you've played Mastermind, how many black key pegs there
 * are.
 *
 * The output of this program is the list of solutions that match all of the 
 * test patterns.
 *
 * The way this source file is laid out is to first declare the libraries that 
 * will be included, then to define the constants that will be used, and then to
 * define the functions that will execute the algorithm.
 *
 * Aaron Weeden, 13 May 2013.
 * aweeden@shodor.org
 */

/*
 * Below are the definitions for the libraries that we want to include.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Below are the definitions of the constants.
 */

/*
 * The constant below defines the number of test patterns.  The programmer must 
 * make sure that this number is equal to the number of lines in the input file.
 * A better approach might be to dynamically determine the number of test 
 * patterns, especially if we wanted to generalize our program to handle 
 * arbitrary input files.
 */
const int NUM_TEST_PATTERNS = 30;

/*
 * The constant below defines how long the string in a test pattern is.  This is
 * also how many characters there are in a solution.  It is also the number of 
 * "positions" (hence the name) that can contain a letter of the solution.
 * Again, a better approach might be to dynamically determine the number of
 * positions, especially if we wanted to generalize our program to handle
 * arbitrary input files.
 */
const int NUM_POSITIONS = 15;

/*
 * The constant below defines how many letters there are in the alphabet of the 
 * language that is used to form a solution.  An improvement of this program 
 * might be to allow the user to choose how many letters are in the solution,
 * and perhaps even to specify what those letters are, perhaps via a 
 * command-line option like -l.
 */
const int NUM_LETTERS = 26;

/*
 * This is the smallest letter than can occur in the alphabet of the language
 * used to form a solution.  Adding to the better approach mentioned for the
 * variable above, the program could perhaps dynamically determine what the
 * smallest letter is based on the letters provided as input.
 */
const char SMALLEST_LETTER = 'a';

/*
 * The constant below defines where the input file is located by specifying its 
 * filename.  The programmer must make sure the file has this name.  A better
 * approach might be to allow the user to specify a different input file,
 * perhaps via a command-line option like -i.
 */
const char INPUT_FILE_NAME[17] = "testPatterns.txt";

/*
 * The constant below defines a letter to use if we are unsure of the value of a
 * character at a particular position.  The programmer must ensure this letter
 * is not part of the character set that can make up a solution.  A better
 * approach may be to find a letter to use programmatically.
 */
const char UNKNOWN = '?';

/*
 * The constants below are used to specify that something is true or false in
 * the Boolean sense.
 */
const int TRUE = 1;
const int FALSE = 0;

/*
 * Below are definitions for functions that are used to execute the algorithm.
 */

/*
 * The function below is used to recursively find all of the solutions after we 
 * have narrowed down certain letters than cannot occur in certain positions.
 */
void findSolutions(char TEST_PATTERNS[NUM_TEST_PATTERNS][NUM_POSITIONS + 1],
        char assumedLetters[NUM_POSITIONS],
        int cannotOccur[NUM_POSITIONS][NUM_LETTERS],
        int numsUncheckedPositions[NUM_TEST_PATTERNS],
        int numsCorrect[NUM_TEST_PATTERNS],
        int lastPositionsChecked[NUM_TEST_PATTERNS]);

/*
 * The main function below serves a few purposes:
 * - It is the entry-point function for the program.
 * - It reads in the input file and stores its information in arrays.
 * - It programmatically figures out certain letters than cannot occur at 
 *   certain positions.
 * - It starts off the function defined above, which recursively finds all the 
 *   solutions.
 *
 * The structure of the source for this function is to first declare the 
 * variables that will be used in the function, and then to use them to execute 
 * the algorithm.
 */
int main() {

    /*
     * The variables are defined below.
     */

    /*
     * Much of the algorithm involves assuming that particular letters are
     * correct in their given positions and then finding solutions that could 
     * exist with those assumptions in mind.  We use the array below to store 
     * letters that we are assuming to be correct and in their correct
     * positions.  For example, if we were to assume that the first letter 
     * is correctly a 'd', then we would store the character 'd' in the array at
     * position 0.
     */
    char assumedLetters[NUM_POSITIONS];

    /*
     * Much of the algorithm also involves proving that particular letters
     * cannot occur in particular positions.  We use the array of arrays below 
     * to store whether a given letter at a given position has been proven 
     * unable to occur.  For example, if we have determined the first letter
     * cannot be an 'x', we store TRUE in position 0 for the outer array and 
     * position 'x' - 'a' for the inner array.  The reason we subtract 'a' from 
     * 'x' is so that we use the integer value of the letter x relative 
     * to the letter a, rather than the actual ASCII value of x.  Note that this
     * is assuming the input file is in all lower-case letters.  A better method
     * may be to first convert the characters of the input file into lower case 
     * before making this assumption.
     */
    int cannotOccur[NUM_POSITIONS][NUM_LETTERS];

    /*
     * The array below will hold a string for each test pattern, not including
     * the count of correct letters in correct positions.  The reason why the 
     * inner array includes a "+ 1" here is because the string must be
     * terminated by one extra character, the null character.
     */
    char TEST_PATTERNS[NUM_TEST_PATTERNS][NUM_POSITIONS + 1];

    /*
     * The array below will hold the counts of correct letters in their correct 
     * positions.  Again, these are the numbers that appear after the strings of
     * letters in the input file.
     */
    int numsCorrect[NUM_TEST_PATTERNS];

    /*
     * The array below will hold the last position we checked in each test 
     * pattern.  This array is used to ensure that we don't check a combination
     * of positions more than once.
     */
    int lastPositionsChecked[NUM_TEST_PATTERNS];

    /*
     * We use the file pointer below for reading in data from the input file.
     */
    FILE *inFile;

    /*
     * The variable below specifies how many characters a line of the input file
     * will be.
     * This is calculated as the length of a test pattern, + 1 for the space in
     * between the letters and the count of correct letters in the correct
     * position, + 1 for that count, + 1 for the
     * newline character, and + 1 for the null terminator.  Note that this
     * assumes that each count of correct characters in the correct position has
     * exactly 1 digit.  A more general approach might be to increase this
     * number to allow for > 10 correct letters in the correct position.
     */
    int FILE_LINE_SIZE = NUM_POSITIONS + 1 + 1 + 1 + 1;

    /*
     * We can use this variable to store the current line of the input file, as
     * we read that file in.
     */
    char line[FILE_LINE_SIZE];

    /*
     * We can use the variables below for loop control as we loop through test 
     * patterns, positions, and letters, respectively.
     */
    int testPatternIdx;
    int positionIdx;
    int letterIdx;

    /*
     * The definition of the algorithm is below.
     */

    /*
     * The first thing we want to do is read in the test pattern data from the 
     * input file.
     */
    /* open the input file for reading. */
    inFile = fopen(INPUT_FILE_NAME, "r");
    /* if the input file was not found, */
    if (inFile == NULL) {
        /* print an error message to the standard error stream. */
        fprintf(stderr, "Error: Cannot find input file \"%s\", exiting.\n",
                INPUT_FILE_NAME);
        /* exit; we cannot continue. */
        exit(EXIT_FAILURE);
    }
    /* keep reading lines until there are none left. */
    for (testPatternIdx = 0; fgets(line, FILE_LINE_SIZE, inFile) != NULL;
            testPatternIdx++) {
        /* parse each line into a string (the test pattern) and a number
         * (the count of correct letters in the correct position). */
        sscanf(line, "%s %d", TEST_PATTERNS[testPatternIdx],
                &numsCorrect[testPatternIdx]);
    }
    /* close the input file. */
    fclose(inFile);

    /*
     * Next we want to initialize the arrays that keep track of which letters
     * we are assuming to be correct and which letters have been proven unable
     * to occur in certain positions.
     */
    /* loop over each position. */
    for (positionIdx = 0; positionIdx < NUM_POSITIONS; positionIdx++) {
        /* indicate that the letter at each position is not assumed yet. */
        assumedLetters[positionIdx] = UNKNOWN;
        /* loop over each letter at each position. */
        for (letterIdx = 0; letterIdx < NUM_LETTERS; letterIdx++) {
            /* indicate that each letter is able to occur at each position. */
            cannotOccur[positionIdx][letterIdx] = FALSE;
        }
    }

    /*
     * We can now make an initial list of some characters that can't occur at
     * some positions by looking at the test patterns that have 0 correct
     * characters in the correct positions.  Each character in these represents 
     * a letter that cannot occur at the given position, otherwise the count of
     * correct letters in correct positions would be greater than 0.
     *
     * We also want to indicate that we have not checked any letters yet for
     * making assumptions on their correctness.
     */
    /* loop over each test pattern. */
    for (testPatternIdx = 0; testPatternIdx < NUM_TEST_PATTERNS;
            testPatternIdx++) {
        /* indicate that we have not checked any letters in the given test
         * pattern to assume their correctness yet. */
        lastPositionsChecked[testPatternIdx] = -1;

        /* if the test pattern has 0 correct letters in the correct position, */
        if (numsCorrect[testPatternIdx] == 0) {
            /* loop over each position for the test pattern. */
            for (positionIdx = 0; positionIdx < NUM_POSITIONS; positionIdx++) {
                /* indicate that the letter in the test pattern at the given 
                 * position cannot occur in a solution. */
                cannotOccur[positionIdx][TEST_PATTERNS[testPatternIdx][
                    positionIdx] - SMALLEST_LETTER] = TRUE;
            }
        }
    }

    /*
     * Now we can start recursively finding solutions based on the letters that 
     * are left.  This function is defined further down in the file.
     */
    findSolutions(TEST_PATTERNS, assumedLetters, cannotOccur, numsCorrect,
            numsCorrect, lastPositionsChecked);

    /*
     * Once the function above has finished executing, the algorithm is
     * finished and we can exit.
     */
    return 0;
}

/*
 * This function is used to recursively find the solutions that match all of
 * the test patterns.
 *
 * It does this not by a brute force search of each possible solution, but
 * rather by looking only at the letters that exist in the test patterns, and
 * recursively picking combinations of letters that match all of the test
 * patterns in terms of the number of correct letters in the correct position.
 *
 * This function takes as input:
 * - the list of test patterns
 * - the list of positions filled with the letters that we are currently
 *   assuming to be correct.  Only one letter per position can be assumed to be
 *   correct at any given time.
 * - the list of letters that cannot occur at each position.  The letters in
 *   this list must be mutually exclusive from the letters in the list above.
 * - the list of numbers that count how many positions still need to be
 *   checked for each test pattern in order to meet the number of correct
 *   letters in the correct position for that test pattern.
 * - the list of numbers that count how many letters are correct and in the
 *   correct position for each test pattern.
 * - the list of numbers that indicate which position we last checked in each
 *   test pattern.  This list prevents us from checking a given combination of
 *   letters in test patterns more than once.
 *
 * As solutions are found, this function prints them to the standard output
 * stream.
 *
 * This function is organized by first declaring the variables needed to execute
 * the algorithm, and then using those variables to execute the algorithm.
 */
void findSolutions(char TEST_PATTERNS[NUM_TEST_PATTERNS][NUM_POSITIONS + 1],
        char assumedLetters[NUM_POSITIONS],
        int cannotOccur[NUM_POSITIONS][NUM_LETTERS],
        int numsUncheckedPositions[NUM_TEST_PATTERNS],
        int numsCorrect[NUM_TEST_PATTERNS],
        int lastPositionsChecked[NUM_TEST_PATTERNS]) {

    /*
     * The variables are defined below.
     */

    /*
     * We can use the variable below to indicate whether or not we found a
     * test pattern during a search.  We can use the value TRUE to indicate we 
     * found one, and FALSE to indicate we did not.
     */
    int foundTestPattern;

    /*
     * We can use the variables below for loop control as we loop through test 
     * patterns, positions, and letters.
     */
    int testPatternIdx;
    int otherTestPatternIdx;
    int positionIdx;
    int otherPositionIdx;
    int letterIdx;

    /*
     * Each time we make a recursion, we want to make copies of arrays before we
     * make any changes, otherwise we will lose the data at the point we
     * started the recursion.  The arrays below are used to make copies.  The
     * data is changed in these arrays before being passed to the recursive
     * function call.
     */
    char newAssumedLetters[NUM_POSITIONS];
    int newCannotOccur[NUM_POSITIONS][NUM_LETTERS];
    int newNumsUncheckedPositions[NUM_TEST_PATTERNS];
    int newLastPositionsChecked[NUM_TEST_PATTERNS];

    /*
     * The algorithm is defined below.
     */

    /*
     * We first look for a test pattern that still has unchecked positions.
     */
    /* indicate we have not found a test pattern with unchecked positions
     * yet. */
    foundTestPattern = FALSE;
    /* loop through each test pattern or until we find a test pattern with
     * unchecked positions. */
    for (testPatternIdx = 0; testPatternIdx < NUM_TEST_PATTERNS &&
            foundTestPattern == FALSE; testPatternIdx++) {
        /* if we find a test pattern with unchecked positions, */
        if (numsUncheckedPositions[testPatternIdx] > 0) {
            /* indicate we found a test pattern with unchecked positions. */
            foundTestPattern = TRUE;
        }
    }

    /*
     * If no test patterns have unchecked positions, then all test patterns
     * have met their required number of correct letters in the correct
     * position, thus we have found a solution (or multiple solutions if any
     * of the positions still has an unknown letter).
     * In this case we want to print the solutions and return.
     */
    /* if we didn't find a test pattern with unchecked positions, */
    if (foundTestPattern == FALSE) {
        /* print the assumed letters -- this is now a solution. */
        printf("%s\n", assumedLetters);
        /* loop over each of the positions. */
        for (positionIdx = 0; positionIdx < NUM_POSITIONS; positionIdx++) {
            /* if the assumed letter at the position is unknown, */
            if (assumedLetters[positionIdx] == UNKNOWN) {
                /* start printing all of the possible letters that can occur
                 * at the position. */
                printf("\twith possible letters for position %d: ",
                        positionIdx);
                /* loop over each letter in the alphabet. */
                for (letterIdx = 0; letterIdx < NUM_LETTERS; letterIdx++) {
                    /* if the letter can occur at the position, */
                    if (cannotOccur[positionIdx][letterIdx] == FALSE) {
                        /* print the letter. */
                        printf("%c", letterIdx + SMALLEST_LETTER);
                    }
                }
                /* print a newline for the next solution. */
                printf("\n");
            }
        }
        /* otherwise, if we found a test pattern with unchecked positions, */
    } else {
        /* because of how we exited the loop, the index of the test pattern we
         * found is 1 less than what it is currently, so decrement. */
        testPatternIdx--;
        /* loop over each of the positions, started with the position after the 
         * one we last checked in the test pattern. */
        for (positionIdx = lastPositionsChecked[testPatternIdx] + 1;
                positionIdx < NUM_POSITIONS; positionIdx++) {
            /* if we haven't assumed anything about the letter at the position, 
             * and the letter in the test pattern at the position has not been
             * proven unable to occur at that position, */
            if (assumedLetters[positionIdx] == UNKNOWN &&
                    cannotOccur[positionIdx][TEST_PATTERNS[testPatternIdx][
                    positionIdx] - SMALLEST_LETTER] == FALSE) {
                /* loop over the positions again, */
                for (otherPositionIdx = 0; otherPositionIdx < NUM_POSITIONS;
                        otherPositionIdx++) {
                    /* copy the assumed letter at the position into a new array
                     * that we will use for recursion. */
                    newAssumedLetters[otherPositionIdx] =
                        assumedLetters[otherPositionIdx];
                    /* loop over each of the letters. */
                    for (letterIdx = 0; letterIdx < NUM_LETTERS; letterIdx++) {
                        /* copy whether the letter can occur at the position
                         * into a new array that we will use for recursion. */
                        newCannotOccur[otherPositionIdx][letterIdx] =
                            cannotOccur[otherPositionIdx][letterIdx];
                    }
                }

                /* loop over the test patterns. */
                for (otherTestPatternIdx = 0;
                        otherTestPatternIdx < NUM_TEST_PATTERNS;
                        otherTestPatternIdx++) {
                    /* copy the number of unchecked positions for the test
                     * pattern into a new array that we will use for
                     * recursion. */
                    newNumsUncheckedPositions[otherTestPatternIdx] =
                        numsUncheckedPositions[otherTestPatternIdx];
                    /* copy the last positions checked for the test pattern
                     * into a new array that we will use for recursion. */
                    newLastPositionsChecked[otherTestPatternIdx] =
                        lastPositionsChecked[otherTestPatternIdx];
                }

                /* indicate that the current position is the last one we
                 * checked in the test pattern that we found. */
                newLastPositionsChecked[testPatternIdx] = positionIdx;
                /* assume that the letter at the current position in the
                 * test pattern we found is correct in the correct position. */
                newAssumedLetters[positionIdx] =
                    TEST_PATTERNS[testPatternIdx][positionIdx];

                /* loop over the test patterns. */
                for (otherTestPatternIdx = 0;
                        otherTestPatternIdx < NUM_TEST_PATTERNS;
                        otherTestPatternIdx++) {
                    /* if the test pattern has the letter we just assumed to be
                     * correct at the given position, */
                    if (TEST_PATTERNS[otherTestPatternIdx][positionIdx] ==
                            newAssumedLetters[positionIdx]) {
                        /* decrement the number of unchecked positions for the
                         * test pattern. */
                        newNumsUncheckedPositions[otherTestPatternIdx]--;
                        /* if the test pattern now has 0 unchecked positions, */
                        if (newNumsUncheckedPositions[otherTestPatternIdx] == 
                                0) {
                            /* loop over all the positions. */
                            for (otherPositionIdx = 0;
                                    otherPositionIdx < NUM_POSITIONS;
                                    otherPositionIdx++) {
                                /* indicate that the letter at the given
                                 * position cannot occur.  This is because we
                                 * have already found all of the letters that
                                 * are correct and in the correct position for
                                 * this test pattern. */
                                newCannotOccur[otherPositionIdx][
                                    TEST_PATTERNS[otherTestPatternIdx][
                                    otherPositionIdx] - SMALLEST_LETTER] = TRUE;
                            }
                        }
                    }
                }
                /*
                 * Recur!
                 */
                findSolutions(TEST_PATTERNS, newAssumedLetters, newCannotOccur,
                        newNumsUncheckedPositions, numsCorrect,
                        newLastPositionsChecked);
            }
        }
    }
}
